/* S1 cpu simulator
   Computer Architecture 1997

   supporting function

   P. Chongstitvatana
   Department of Computer Engineering
   Chulalongkorn University

   include clock counting and cpi calculation 15 Aug 97
   include showms readmpgm  printout 	      17 July 98
*/

int itime[] = { 5,5,3,3, 0,0,0,  /* load,store,jmp,call,-,-,- */
		4,5,5,4,         /* mov, mov (r), mov r (r), add, */
		3,4,7,7,7 };     /* cmp,inc,ret,push,pop */
int clock, ninst;

countclock()
{
  int inst;
  if( IRop() == IXOP )
    inst = IXOP + IRxop();
  else
    inst = IRop();
  clock += itime[inst];
  ninst++;
}

report()
{
  float cpi;
  if ( ninst > 0 ) {
    cpi = ((float) clock)/ninst;
    printf("%d instructions %d clocks CPI = %f\n",
	 ninst,clock,cpi);
  }
}

/* IR decode */
IRop() { return( ( Ir & 0xE000 ) >> 13 ); } /* bit 15..13 */
IRr()  { return( ( Ir & 0x1C00 ) >> 10 ); } /* bit 12..10 */
IRads(){ return(   Ir & 0x03FF         ); } /* bit 9..0   */
IRxop(){ return( ( Ir & 0x1E00 ) >> 9  ); } /* bit 12..9  */
IRr1() { return( ( Ir & 0x0038 ) >> 3  ); } /* bit 5..3   */
IRr2() { return(   Ir & 0x0007         ); } /* bit 2..0   */

stop()
{
  if( core ) dump(start,nbyte);
  report();
  exit(0);
}

trap( int code )	/* special function */
{
  switch( code ) {
  case 0 : stop(); break;
  case 1 : printf("%d\n",R[0]); mpc = UIFETCH; break;
  default : printf("undef opcode\n"); stop();
  }
}

/* load object code from file */

geta(FILE *fp)	/* get one word, return first ch */
{
  char s[32];
  fscanf(fp,"%s",s);
  return s[0] ;
}

geti(FILE *fp)	/* get one integer */
{
  int c;
  fscanf(fp,"%d",&c);
  return c;
}

/* set maximum of number to be read to 1000
   to prevent infinite loop when input is malformed
   the input file name is fixed to "in.obj"
*/
loadprogram()
{
  int cnt,ip,c,c1,c2;
  FILE *fp;

  fp = fopen("in.obj","r");
  ip = 0;
  for(cnt=0; cnt<1000; cnt++ ) {
    switch( geta(fp) ) {
    case 'a': ip = geti(fp);		/* set address */
	break;
    case 'i': c = geti(fp);		/* instruction */
	c1 = geti(fp);
	c2 = geti(fp);
	if(c == IXOP)
	  M[ip] = (c << 13) | (c1 << 9) | (c2 << 3) | geti(fp);
	else
	  M[ip] = (c << 13) | (c1 << 10) | c2;
	ip++;
	break;
    case 'w': M[ip] = geti(fp);		/* set constant data */
	ip++;
	break;
    case 't': trace = 1;		/* trace mode */
	break;
    case 'd': core = 1;                 /* dump start nbyte */
	start = geti(fp);
	nbyte = geti(fp);
	break;
    case 'e': cnt = 1000;		/* end */
	break;
    }
  }
  fclose(fp);
  printf("load program, last address %d\n",ip);
}

dump(int ads, int n)  /* dump n words starting from ads */
{
  int i;
  for(i=ads; i<ads+n; i++)
    printf("%4d %4x\n",i,M[i]);
}

show()		/* show some registers */
{
  int op, ox;
  op = IRop();
  if( op == IXOP ) ox = IRxop(); else ox = 0;
  printf("PC %4d IR %4x op %d xop %d R0 %d R1 %d R2 %d R7 %d Z %d S %d\n",
	Pc,Ir,op,ox,R[0],R[1],R[2],R[7],Z,S);
}

showms()	/* show micro step */
{
  char d[80];
  if( IRop() == IXOP )
    sprintf(d,"x:%d:%d:%d",IRxop(),IRr1(),IRr2());
  else
    sprintf(d,"%d:%d:%d",IRop(),IRr(),IRads());
  printf("mpc %d ir %s pc %d mar %d mdr %d\n",mpc,d,Pc,Mar,Mdr);
}

printout()  /* print microprogram for visual check */
{
  int i,j;
  for( i=0;i<MXMS ;i++ ) {
    printf("%2d  ",i);
    for( j=0; j<MSW; j++ ) {
      printf("%c",ms[i][j]);
      if( j == 4  || j == 10 || j == 14 | j == 18 ||
	  j == 20 || j == 21 || j == 25 )
	  printf(" ");
    }
    printf("\n");
  }
}

readmpgm()	/* read microprogram mpgm.txt */
{
  int i,n;
  FILE *fp;
  char ibuf[MXSTR];

  fp = fopen("mpgm.txt","r");
  fgets(ibuf,MXSTR,fp);
  sscanf(ibuf,"%d",&n);	  /* length of mpgm */
  for( i=0; i<n; i++ ) {
    fgets(ibuf,MXSTR,fp);
    strncpy(ms[i],ibuf+4,MSW);
  }
  fclose(fp);
}
